@@ -3,6 +3,7 @@ import functools
 import logging
 import sys
 from contextlib import suppress
+from urllib.parse import quote
 
 import backoff
 import httpx
@@ -159,7 +160,7 @@ class B2(Backend):
         upload_url, upload_token = await self._get_upload_url_token()
         upload_headers = {
             'authorization': upload_token,
-            'x-bz-file-name': name,
+            'x-bz-file-name': quote(name),
             'content-type': 'application/octet-stream',
             'content-length': str(len(data)),
             'x-bz-content-sha1': 'do_not_verify',  # TODO
@@ -172,7 +173,7 @@ class B2(Backend):
         upload_url, upload_token = await self._get_upload_url_token()
         upload_headers = {
             'authorization': upload_token,
-            'x-bz-file-name': name,
+            'x-bz-file-name': quote(name),
             'content-type': 'application/octet-stream',
             'content-length': str(length),
             'x-bz-content-sha1': 'do_not_verify',  # TODO
@@ -194,6 +195,26 @@ class B2(Backend):
         response = await self._client.get(url, headers=headers)
         return await response.aread()
 
+    @utils.requires_auth
+    @backoff_reauth
+    async def download_stream(self, name, stream):
+        bucket = await self._get_bucket()
+        url = f'{self._auth.downloadUrl}/file/{bucket.name}/{name}'
+        headers = {'authorization': self._auth.authorizationToken}
+
+        async with self._client.stream('GET', url, headers=headers) as response:
+            content_length = response.headers.get('content-length')
+            if content_length is not None:
+                content_length = int(content_length)
+
+            try:
+                stream.truncate(content_length)
+                async for chunk in response.aiter_bytes(128_000):
+                    stream.write(chunk)
+            except:
+                stream.seek(0)
+                raise
+
     @utils.requires_auth
     @backoff_reauth
     async def _list_file_names(self, *, start_file_name=None, prefix=''):
@@ -236,7 +257,21 @@ class B2(Backend):
         bucket = await self._get_bucket()
         params = {'bucketId': bucket.id, 'fileName': name}
         headers = {'authorization': self._auth.authorizationToken}
-        await self._client.post(url, json=params, headers=headers)
+
+        try:
+            await self._client.post(url, json=params, headers=headers)
+        except httpx.HTTPStatusError as e:
+            if e.response.status_code == httpx.codes.BAD_REQUEST:
+                # NOTE: b2_hide_file documentation lies about the response code.
+                # Also, those codes are not documented
+                try:
+                    decoded_error = e.response.json()
+                except ValueError:
+                    pass
+                else:
+                    if decoded_error.get('code') in ('already_hidden', 'no_such_file'):
+                        return
+            raise
 
     async def close(self):
         with suppress(AttributeError):
