@@ -1,8 +1,9 @@
+import contextlib
 import hashlib
 import hmac
 import logging
 from datetime import datetime
-from urllib.parse import urlencode
+from urllib.parse import quote, urlencode
 from xml.etree.ElementTree import XMLPullParser
 
 import backoff
@@ -122,7 +123,7 @@ class S3Compatible(Backend, short_name='S3C'):
             timeout=None, event_hooks={'response': [_raise_for_status_hook]}
         )
 
-    async def _make_request(
+    def _prepare_request(
         self,
         method,
         canonical_uri,
@@ -132,7 +133,10 @@ class S3Compatible(Backend, short_name='S3C'):
         headers=None,
         **kwargs,
     ):
-        url = self.url + canonical_uri
+        # HTTPX encodes strings before the request, but we need to know
+        # the actual value for signature computation
+        encoded_canonical_uri = quote(canonical_uri)
+        url = self.url + encoded_canonical_uri
         if query:
             query_string = urlencode(sorted(query.items()))
             url += f'?{query_string}'
@@ -154,7 +158,7 @@ class S3Compatible(Backend, short_name='S3C'):
         signed_headers = ";".join(canonical_headers)
         canonical_request = _make_canonical_request(
             method=method,
-            canonical_uri=canonical_uri,
+            canonical_uri=encoded_canonical_uri,
             canonical_query=query_string,
             canonical_headers=_make_canonical_headers(canonical_headers),
             signed_headers=signed_headers,
@@ -181,7 +185,52 @@ class S3Compatible(Backend, short_name='S3C'):
         headers['x-amz-content-sha256'] = payload_digest
         headers['x-amz-date'] = x_amz_date
         headers['authorization'] = authorization_header
-        return await self._client.request(method, url, headers=headers, **kwargs)
+        return self._client.build_request(method, url, headers=headers, **kwargs)
+
+    async def _make_request(
+        self,
+        method,
+        canonical_uri,
+        *,
+        query=None,
+        payload_digest,
+        headers=None,
+        **kwargs,
+    ):
+        request = self._prepare_request(
+            method,
+            canonical_uri,
+            query=query,
+            payload_digest=payload_digest,
+            headers=headers,
+            **kwargs,
+        )
+        return await self._client.send(request)
+
+    @contextlib.asynccontextmanager
+    async def _make_streaming_request(
+        self,
+        method,
+        canonical_uri,
+        *,
+        query=None,
+        payload_digest,
+        headers=None,
+        **kwargs,
+    ):
+        request = self._prepare_request(
+            method,
+            canonical_uri,
+            query=query,
+            payload_digest=payload_digest,
+            headers=headers,
+            **kwargs,
+        )
+        response = await self._client.send(request, stream=True)
+        try:
+            yield response
+        finally:
+            await response.aclose()
 
     @backoff_on_httperror
     async def exists(self, name):
@@ -239,6 +288,25 @@ class S3Compatible(Backend, short_name='S3C'):
         )
         return await response.aread()
 
+    @backoff_on_httperror
+    async def download_stream(self, name, stream):
+        async with self._make_streaming_request(
+            'GET',
+            f'/{self.bucket_name}/{name}',
+            payload_digest=_empty_payload_digest,
+        ) as response:
+            content_length = response.headers.get('content-length')
+            if content_length is not None:
+                content_length = int(content_length)
+
+            try:
+                stream.truncate(content_length)
+                async for chunk in response.aiter_bytes(128_000):
+                    stream.write(chunk)
+            except:
+                stream.seek(0)
+                raise
+
     @backoff_on_httperror
     async def _list_objects(self, *, continuation_token=None, prefix=''):
         query = {'list-type': '2'}
