@@ -38,6 +38,84 @@ class TestHelperMethods:
         assert name == 'GHIJKLMNOPQR'
         assert tag == '0123456789ABCDEF'
 
+    def test_read_metadata_path(self, local_repo, tmp_path):
+        file = tmp_path / 'some_file'
+        with file.open('wb'):
+            pass
+
+        with patch.object(os, 'stat') as stat_mock:
+            metadata = local_repo.read_metadata(file)
+
+        stat_result = stat_mock.return_value
+        expected_metadata = {
+            'st_mode': stat_result.st_mode,
+            'st_uid': stat_result.st_uid,
+            'st_gid': stat_result.st_gid,
+            'st_size': stat_result.st_size,
+            'st_atime_ns': stat_result.st_atime_ns,
+            'st_mtime_ns': stat_result.st_mtime_ns,
+            'st_ctime_ns': stat_result.st_ctime_ns,
+        }
+        assert metadata == expected_metadata
+
+    def test_read_metadata_fd(self, local_repo, tmp_path):
+        path = tmp_path / 'some_file'
+        with path.open('wb') as file, patch.object(os, 'fstat') as fstat_mock:
+            metadata = local_repo.read_metadata(file.fileno())
+
+        stat_result = fstat_mock.return_value
+        expected_metadata = {
+            'st_mode': stat_result.st_mode,
+            'st_uid': stat_result.st_uid,
+            'st_gid': stat_result.st_gid,
+            'st_size': stat_result.st_size,
+            'st_atime_ns': stat_result.st_atime_ns,
+            'st_mtime_ns': stat_result.st_mtime_ns,
+            'st_ctime_ns': stat_result.st_ctime_ns,
+        }
+        assert metadata == expected_metadata
+
+    def test_restore_metadata(self, local_repo, tmp_path):
+        file = tmp_path / 'some_file'
+        with file.open('wb'):
+            pass
+
+        metadata = {
+            'st_mode': 12345,
+            'st_uid': 67890,
+            'st_gid': 54321,
+            'st_size': 98765,
+            'st_atime_ns': 123,
+            'st_mtime_ns': 321,
+            'st_ctime_ns': 987,
+        }
+
+        # We don't know the timestamp resolution of this system and frankly we don't
+        # care. If os.utime gets called with correct arguments, it's enough
+        with patch.object(os, 'utime') as utime_mock:
+            local_repo.restore_metadata(file, metadata)
+
+        utime_mock.assert_called_once_with(file, ns=(123, 321))
+
+    def test_restore_metadata_legacy_timestamps(self, local_repo, tmp_path):
+        file = tmp_path / 'some_file'
+        with file.open('wb'):
+            pass
+
+        metadata = {
+            'st_mode': 12345,
+            'st_uid': 67890,
+            'st_gid': 54321,
+            'st_size': 98765,
+            'st_atime': 123,
+            'st_mtime': 321,
+            'st_ctime': 987,
+        }
+        with patch.object(os, 'utime') as utime_mock:
+            local_repo.restore_metadata(file, metadata)
+
+        utime_mock.assert_called_once_with(file, times=(123, 321))
+
 
 class TestInit:
     @pytest.mark.asyncio
@@ -595,10 +673,9 @@ class TestRestore:
         second_file.parent.mkdir(exist_ok=True, parents=True)
         second_file.write_bytes(second_data)
 
-        snapshot_params = await local_repo.snapshot(paths=[first_file, second_file])
-        result = await local_repo.restore(
-            snapshot_regex=snapshot_params.name, path=tmp_path
-        )
+        snapshot = await local_repo.snapshot(paths=[first_file, second_file])
+        result = await local_repo.restore(snapshot_regex=snapshot.name, path=tmp_path)
+
         assert set(result.files) == {str(first_file), str(second_file)}
         assert tmp_path.joinpath(*first_file.parts[1:]).read_bytes() == first_data
         assert tmp_path.joinpath(*second_file.parts[1:]).read_bytes() == second_data
@@ -658,7 +735,7 @@ class TestRestore:
         assert tmp_path.joinpath(*file.parts[1:]).read_bytes() == second_version
 
     @pytest.mark.asyncio
-    async def test_snapshot_version(self, local_repo, tmp_path):
+    async def test_specific_snapshot_version(self, local_repo, tmp_path):
         await local_repo.init(
             settings={
                 'encryption': None,
@@ -719,7 +796,7 @@ class TestDeleteSnapshots:
         second_snapshot_paths = [
             local_repo._chunk_digest_to_location(x) for x in second_snapshot.chunks
         ]
-        await local_repo.delete_snapshots([second_snapshot.name])
+        await local_repo.delete_snapshots([second_snapshot.name], confirm=False)
 
         assert all(map(local_backend.exists, first_snapshot_paths))
         assert not any(map(local_backend.exists, second_snapshot_paths))
@@ -761,7 +838,7 @@ class TestDeleteSnapshots:
         second_snapshot_paths = [
             local_repo._chunk_digest_to_location(x) for x in second_snapshot.chunks
         ]
-        await local_repo.delete_snapshots([second_snapshot.name])
+        await local_repo.delete_snapshots([second_snapshot.name], confirm=False)
 
         assert all(map(local_backend.exists, first_snapshot_paths))
         assert not any(map(local_backend.exists, second_snapshot_paths))
@@ -802,7 +879,7 @@ class TestDeleteSnapshots:
         second_snapshot_paths = [
             local_repo._chunk_digest_to_location(x) for x in second_snapshot.chunks
         ]
-        await local_repo.delete_snapshots([second_snapshot.name])
+        await local_repo.delete_snapshots([second_snapshot.name], confirm=False)
 
         assert all(map(local_backend.exists, first_snapshot_paths))
         assert all(map(local_backend.exists, second_snapshot_paths))
@@ -846,7 +923,7 @@ class TestDeleteSnapshots:
             local_repo._chunk_digest_to_location(x) for x in second_snapshot.chunks
         ]
 
-        await local_repo.delete_snapshots([second_snapshot.name])
+        await local_repo.delete_snapshots([second_snapshot.name], confirm=False)
 
         assert not any(map(local_backend.exists, first_snapshot_paths))
         assert not any(map(local_backend.exists, second_snapshot_paths))
@@ -871,7 +948,7 @@ class TestDeleteSnapshots:
             local_repo._chunk_digest_to_location(x) for x in snapshot.chunks
         ]
         await local_repo.snapshot(paths=[file])
-        await local_repo.delete_snapshots([snapshot.name])
+        await local_repo.delete_snapshots([snapshot.name], confirm=False)
         assert all(map(local_backend.exists, snapshot_paths))
         assert not local_backend.exists(snapshot.location)
 
@@ -894,7 +971,7 @@ class TestDeleteSnapshots:
             local_repo._chunk_digest_to_location(x) for x in snapshot.chunks
         ]
 
-        await local_repo.delete_snapshots([snapshot.name])
+        await local_repo.delete_snapshots([snapshot.name], confirm=False)
 
         assert not any(map(local_backend.exists, snapshot_paths))
         assert not local_backend.exists(snapshot.location)
@@ -1098,7 +1175,7 @@ class TestClean:
         assert all(map(local_backend.exists, chunks_paths))
 
 
-class TestUpload:
+class TestUploadObjects:
     @pytest.fixture(autouse=True)
     def change_cwd(self, tmp_path):
         before = os.getcwd()
@@ -1121,7 +1198,7 @@ class TestUpload:
             path.parent.mkdir(parents=True, exist_ok=True)
             path.write_bytes(data)
 
-        await repository.upload(
+        await repository.upload_objects(
             [
                 files_base_path / 'file',
                 files_base_path / 'directory',
@@ -1144,7 +1221,7 @@ class TestUpload:
         )
 
     @pytest.mark.asyncio
-    async def test_overwrites(self, local_backend, tmp_path):
+    async def test_overwrites(self, tmp_path):
         backend = Local(tmp_path / 'backend')
         backend.upload('files/file', b'<old data>')
 
@@ -1154,12 +1231,12 @@ class TestUpload:
 
         repository = Repository(backend, concurrent=5)
 
-        await repository.upload([files_base_path / 'file'])
+        await repository.upload_objects([files_base_path / 'file'])
         assert set(backend.list_files()) == {'files/file'}
         assert backend.download('files/file') == b'<updated data>'
 
     @pytest.mark.asyncio
-    async def test_skip_existing(self, local_backend, tmp_path):
+    async def test_skip_existing(self, tmp_path):
         backend = Local(tmp_path / 'backend')
         backend.upload('files/file', b'<old data>')
 
@@ -1169,6 +1246,103 @@ class TestUpload:
 
         repository = Repository(backend, concurrent=5)
 
-        await repository.upload([files_base_path / 'file'], skip_existing=True)
+        await repository.upload_objects([files_base_path / 'file'], skip_existing=True)
         assert set(backend.list_files()) == {'files/file'}
         assert backend.download('files/file') == b'<old data>'
+
+
+class TestDownloadObjects:
+    @pytest.fixture(autouse=True)
+    def populate_backend(self, local_backend):
+        local_backend.upload('file', b'<backend data>')
+        local_backend.upload('nested/file', b'<nested backend data>')
+        local_backend.upload(
+            'very/very/nested/file', b'<very very nested backend data>'
+        )
+
+    @pytest.mark.asyncio
+    async def test_overwrite(self, local_backend, local_repo, tmp_path):
+        target_path = tmp_path / 'downloaded_files'
+        (target_path / 'very/very/nested').mkdir(parents=True)
+        (target_path / 'file').write_bytes(b'<local data>')
+        (target_path / 'very/very/nested/file').write_bytes(
+            b'<very very nested local data>'
+        )
+
+        await local_repo.download_objects(path=target_path)
+        assert (target_path / 'file').read_bytes() == b'<backend data>'
+        assert (target_path / 'nested/file').read_bytes() == b'<nested backend data>'
+        assert (
+            target_path / 'very/very/nested/file'
+        ).read_bytes() == b'<very very nested backend data>'
+
+    @pytest.mark.asyncio
+    async def test_skip_existing(self, local_backend, local_repo, tmp_path):
+        target_path = tmp_path / 'downloaded_files'
+        (target_path / 'very/very/nested').mkdir(parents=True)
+        (target_path / 'file').write_bytes(b'<local data>')
+        (target_path / 'very/very/nested/file').write_bytes(
+            b'<very very nested local data>'
+        )
+
+        await local_repo.download_objects(path=target_path, skip_existing=True)
+        assert (target_path / 'file').read_bytes() == b'<local data>'
+        assert (target_path / 'nested/file').read_bytes() == b'<nested backend data>'
+        assert (
+            target_path / 'very/very/nested/file'
+        ).read_bytes() == b'<very very nested local data>'
+
+    @pytest.mark.asyncio
+    async def test_regex_filter(self, local_backend, local_repo, tmp_path):
+        target_path = tmp_path / 'downloaded_files'
+        await local_repo.download_objects(path=target_path, object_regex='^very|^file')
+        assert {x for x in target_path.rglob('*') if x.is_file()} == {
+            target_path / 'file',
+            target_path / 'very/very/nested/file',
+        }
+
+
+class TestListObjects:
+    @pytest.fixture(autouse=True)
+    def populate_backend(self, local_backend):
+        local_backend.upload('file', b'<backend data>')
+        local_backend.upload('nested/file', b'<nested backend data>')
+        local_backend.upload(
+            'very/very/nested/file', b'<very very nested backend data>'
+        )
+
+    @pytest.mark.asyncio
+    async def test_all(self, local_repo):
+        result = await local_repo.list_objects()
+        assert set(result.paths) == {
+            'file',
+            'nested/file',
+            'very/very/nested/file',
+        }
+
+    @pytest.mark.asyncio
+    async def test_regex_filter(self, local_repo):
+        result = await local_repo.list_objects(object_regex='^very|^file')
+        assert set(result.paths) == {
+            'file',
+            'very/very/nested/file',
+        }
+
+
+class TestDeleteObjects:
+    @pytest.fixture(autouse=True)
+    def populate_backend(self, local_backend):
+        local_backend.upload('file', b'<backend data>')
+        local_backend.upload('nested/file', b'<nested backend data>')
+        local_backend.upload(
+            'very/very/nested/file', b'<very very nested backend data>'
+        )
+
+    @pytest.mark.asyncio
+    async def test_ok(self, local_backend, local_repo):
+        await local_repo.delete_objects(
+            ['file', 'very/very/nested/file'], confirm=False
+        )
+        assert not local_backend.exists('file')
+        assert not local_backend.exists('very/very/nested/file')
+        assert local_backend.exists('nested/file')
